/* inverse_brun
 *  Copyright (C) 2000 Luc Brun
 *  email: brun@leri.univ-reims.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include "types.h"
#include "conversion.h"
#include "conv_inter_conv_aux.h"
/*#include "definitions.h"*/
/*#include "macroI1I2I3.h"*/



/****************Espace I1 I2 I3 *******************************************/
double RGB_to_I1(int R,int G,int B)
{
    static float inv_quotient = 1.0/3.0;
    double retour  = (R+G+B)*inv_quotient;
    
    return (retour);
}

double RGB_to_I2(int R,int G,int B)
{
    return(R-B);
}

double RGB_to_I3(int R,int G,int B)
{
  double retour = G-(R+B)*0.5;

  return (retour);
}

/****************************************************************************/

/****************Espace L u v  *******************************************/

double RGB_to_L(int R,int G,int B)
{
  double retour = RGB_to_Luv(R,G,B);

  return(retour);
}

double RGB_to_u(int R,int G,int B)
{
  double X,Y,Z,up,u;
  double L = RGB_to_Luv(R,G,B);
  RGB_to_XYZ(R,G,B,&X,&Y,&Z);
  
  if((X==0.0)&&(Y==0.0)&&(Z==0.0))
    return 0;

  up = 4*X/(X+15*Y+3*Z);
  u = 13*L*(up-Uw);
  return u;
    
}
double  RGB_to_v(int R,int G,int B)
{
  double X,Y,Z,vp,v;
  double L = RGB_to_L(R,G,B);
  
  RGB_to_XYZ(R,G,B,&X,&Y,&Z);
  
  if((X==0.0)&&(Y==0.0)&&(Z==0.0))
    return 0.0;
  
  vp = 9*Y/(X+15*Y+3*Z);
  v = 13*L*(vp-Vw);
  return v;
}   
/****************************************************************************/

/****************Espace Y I Q  *******************************************/
double RGB_to_Y(int R,int G,int B)
{
    static float tab_Y[3] = {0.299,0.587,0.133};
    double Y;

    Y = tab_Y[0]*R+tab_Y[1]*G+tab_Y[2]*B+0.5;

    return(Y);
}
double RGB_to_I(int R,int G,int B)
{
    static float tab_I[3] = {0.596,-0.274,-0.322};
    double I;

    I = tab_I[0]*R+tab_I[1]*G+tab_I[2]*B;

    return(I);
}
double RGB_to_Q(int R,int G,int B)
{
  static float tab_Q[3] = {0.211,-0.253,0.312};
  double Q;
  
  Q = tab_Q[0]*R+tab_Q[1]*G+tab_Q[2]*B;
  
  return(Q);
}
/****************************************************************************/

/****************Espace R G B /***********************************************************/

double RGB_to_R(int R,int G,int B)
{
    return(R);
}
double RGB_to_G(int R,int G,int B)
{
    return(G);
}
double RGB_to_B(int R,int G,int B)
{
    return(B);
}

/****************************************************************************/
/*                            Espace K1K2K3                       */
/****************************************************************************/
double RGB_to_K1(int R,int G,int B)
{
  return(R+G);
}

double RGB_to_K2(int R,int G,int B)
{
  return(R-G);
}
double RGB_to_K3(int R,int G,int B)
{
  return(B-0.5*(R+G));
}

PFconversion* changer_espace(int choix)
{
  static PFconversion conversion[][3]= 
    {
      {RGB_to_R,RGB_to_G,RGB_to_B},
      {RGB_to_I1,RGB_to_I2,RGB_to_I3},
      {RGB_to_L,RGB_to_u,RGB_to_v},
      {RGB_to_Y,RGB_to_I,RGB_to_Q},
      {RGB_to_K1,RGB_to_K2,RGB_to_K3},
    };
  return(conversion[choix]);
}











