/* inverse_brun
 *  Copyright (C) 2000 Luc Brun
 *  email: brun@leri.univ-reims.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include <objet.h>
#include "image_dist.h"
#include <dart_list.h>
#include <tab_mark_dart.h>
#include <tab_mark.h>
#include <type_obj.h>
#include <math.h>

#define Dimx     SELF(dimx)
#define Dimy     SELF(dimy)
#define Tab      SELF(tab)
#define Delaunay SELF(delaunay)
#define Nb_src   SELF(nb_src)
#define Frontier SELF(frontier)

#define NULL_DELAUNAY (dart_list*) NULL

struct dist_elt
{
  float dx;
  float dy;
  int  src_nb;
};

typedef struct dist_elt *dist_elt;

struct image_dist
{
  int dimx;
  int dimy;
  int nb_src;
  dist_elt tab;
  dart_list *delaunay;  
  tab_mark frontier;
};

/* Declaration des fonctions privees */

static void  set_tab_to_zero_priv          (image_dist);

static void insert_dist_to_sources_priv    (image_dist,
					    double*,
					    int);

static void  insert_dist_to_one_source_priv(image_dist,
					    double,
					    double,
					    int);

static void set_distances_priv              (image_dist,
					     bouleen,
					     tab_mark_dart*);

static void traite_ligne_ascendant_priv     (image_dist,
					     dist_elt,
					     dist_elt,
					     int,
					     int);

static void traite_ligne_descendant_priv    (image_dist,
					     dist_elt,
					     dist_elt,
					     int,
					     int,
					     bouleen,
					     int,
					     tab_mark_dart*);

static float give_max_dist_priv             (image_dist);

static void update_neighbor_priv(image_dist,
				 tab_mark_dart*,
				 int,
				 int);

static void set_delaunay_priv(image_dist,tab_mark_dart*);

/* Corps des fonctions. */

image_dist FAIRE_image_dist()
{
  INSTANCIER(image_dist);

  Dimx =0;
  Dimy = 0;
  Tab = NULL;
  Delaunay = NULL_DELAUNAY;

  return self;
}
void DEFAIRE_image_dist(image_dist self)
{
  int i;

  assert(self);
  
  if(Tab != NULL)
    free(Tab);

  if(Delaunay !=  NULL_DELAUNAY)
    for(i=0;i<Nb_src;i++)
      DEFAIRE_dart_list(Delaunay[i]);
  
  free(self);
}

void image_dist_initialize(image_dist self,
			   int dimx,
			   int dimy,
			   double * src,
			   int nb_src,
			   bouleen compute_adjacency)
{
  int i;
  tab_mark_dart *mark;

  assert(self);
  assert(Tab == NULL);

  Nb_src = nb_src;
  Dimx = dimx;
  Dimy = dimy;
  Tab  = (dist_elt) malloc(dimx*dimy*sizeof(struct dist_elt));
  assert(Tab);

  set_tab_to_zero_priv(self);
  
  insert_dist_to_sources_priv(self,
			      src,
			      Nb_src);



  Frontier = FAIRE_tab_mark(Dimy,Dimx);
  Delaunay = (dart_list*) malloc(Nb_src*sizeof(dart_list));
  mark     = (tab_mark_dart*) malloc(Nb_src*sizeof(tab_mark_dart));

  assert(Delaunay);
  assert(mark);
  
  for(i=0;i<Nb_src;i++)
    {
      Delaunay[i] = FAIRE_dart_list();
      mark[i]     = FAIRE_tab_mark_dart();
    }

  set_distances_priv(self,compute_adjacency,mark);

  for(i=0;i<Nb_src;i++)
    DEFAIRE_tab_mark_dart(mark[i]);
}
void image_dist_set_adjacency_graph(image_dist self)
{
  
  int i;
  tab_mark_dart *mark;

  Frontier = FAIRE_tab_mark(Dimy,Dimx);
  Delaunay = (dart_list*) malloc(Nb_src*sizeof(dart_list));
  mark     = (tab_mark_dart*) malloc(Nb_src*sizeof(tab_mark_dart));
  assert(Delaunay);
  assert(mark);
  
  for(i=0;i<Nb_src;i++)
    {
      Delaunay[i] = FAIRE_dart_list();
      mark[i]     = FAIRE_tab_mark_dart();
    }

  set_delaunay_priv(self,mark);
  for(i=0;i<Nb_src;i++)
    DEFAIRE_tab_mark_dart(mark[i]);
}
float  image_dist_mean_cardinal_of_delaunay_graph(image_dist self)
{
  int i;
  float mean_card=0;
  for(i=0;i<Nb_src;i++)
    mean_card += dart_list_give_cardinal(Delaunay[i]);

  mean_card /=((float)Nb_src);
  
  return mean_card;
}
int image_dist_give_cell_nb(image_dist self,int x,int y)
{
  dist_elt cell;

  assert(self);
  assert(0<=x &&  x < Dimx);
  assert(0<=y &&  y < Dimy);
  
  cell = Tab + y*Dimx+x;

  return cell->src_nb;
}
image image_dist_give_image(image_dist self)
{
  float max;
  image img;
  dist_elt tab;
  int pix_value;
  float real_value ;

/*  max = give_max_dist_priv(self);*/

  img = FAIRE_image();
  image_initialize(img,1,Dimx,Dimy);
  image_debut(img);
  tab = Tab;
  do
    {
/*      real_value = sqrt(tab->dx*tab->dx + tab->dy*tab->dy)*(256.0/max);
      pix_value  = (int)real_value;*/
      pix_value = tab->src_nb;
      image_ecrire_pixel(img,&pix_value);
      tab++;
    }
  while(image_pixel_suivant(img));

  return img;

}
dart_list image_dist_give_adjacent_cells(image_dist self,
					 int cell_nb)
{
  assert(self);
  
  return Delaunay[cell_nb];
}
bouleen image_dist_is_point_frontier(image_dist self,
				     int x,
				     int y)
{
  Point p;
  
  COORDX(p) = x;
  COORDY(p) = y;
  
  return tab_mark_is_marked(Frontier,&p);
}

static void set_delaunay_priv(image_dist self, tab_mark_dart* mark)
{
  int i,j;
  dist_elt up,right,tab=Tab;
  Point p,q;
  
  COORDX(p) = 0;
  COORDY(p) = 0;
  for(j=0;j<Dimx-1;j++)
    {
      right = tab+1;
      if(tab->src_nb != right->src_nb)
	{
	  update_neighbor_priv(self,mark,tab->src_nb,right->src_nb);
	  tab_mark_mark_point(Frontier,&p);
	  COORDX(q) = COORDX(p)+1;
	  COORDY(q) = COORDY(p);
	  tab_mark_mark_point(Frontier,&q);
	}
      tab++;
      COORDX(p)++;
    }
  tab++;
  for(i=0;i<Dimy-1;i++)
    {
      COORDX(p)=0;
      COORDY(p)++;
      for(j=0;j<Dimx-1;j++)
	{
	  right = tab+1;
	  up    = tab-Dimx;
	  if(tab->src_nb != right->src_nb)
	    {
	      update_neighbor_priv(self,mark,tab->src_nb,right->src_nb);
	      tab_mark_mark_point(Frontier,&p);
	      COORDX(q) = COORDX(p)+1;
	      COORDY(q) = COORDY(p);
	      tab_mark_mark_point(Frontier,&q);
	    }
	  if(tab->src_nb != up->src_nb)
	    {
	      update_neighbor_priv(self,mark,tab->src_nb,up->src_nb);
	      tab_mark_mark_point(Frontier,&p);
	      COORDX(q) = COORDX(p);
	      COORDY(q) = COORDY(p)-1;
	      tab_mark_mark_point(Frontier,&q);
	    }
	  tab++;
	  COORDX(p)++;
	}  
      tab++;
    }
}
static void update_neighbor_priv(image_dist self,
				 tab_mark_dart* mark,
				 int ind1,
				 int ind2)
{
  if(!tab_mark_dart_is_marked(mark[ind1],ind2))
    {
      tab_mark_dart_mark_dart(mark[ind1],ind2);
      dart_list_insert(Delaunay[ind1],ind2);

    }
  if(!tab_mark_dart_is_marked(mark[ind2],ind1))
    {
      tab_mark_dart_mark_dart(mark[ind2],ind1);
      dart_list_insert(Delaunay[ind2],ind1); 
    }
}
static float give_max_dist_priv(image_dist self)
{
  int i,j;
  float max,real_value;
  dist_elt tab = Tab;

  for(j=0;j<Dimx;j++)
    for(i=0;i<Dimy;i++)
      {
	real_value = sqrt(tab->dx*tab->dx +tab->dy*tab->dy);
	if(max < real_value)
	  max = real_value;	
	tab++;
      }
  return max;
}
static void set_tab_to_zero_priv(image_dist self)
{
  int i,j;
  dist_elt tab_aux;

  tab_aux = Tab;

  for(i=0;i<Dimx;i++)
    for(j=0;j<Dimy;j++)
      {
	tab_aux->dx      = -1.0;
	tab_aux->dy      = -1.0;
	tab_aux->src_nb = -1;
	tab_aux++;
      }
}

static void insert_dist_to_sources_priv(image_dist self,
					double *src,
					int nb_src)
{
  int i;
  double x,y;

  for(i=0;i<2*nb_src;i+=2)
    {
      x = src[i];
      y = src[i+1];
      insert_dist_to_one_source_priv(self,x,y,(i/2));
    }
}
static void  insert_dist_to_one_source_priv(image_dist self,
					    double x,
					    double y,
					    int src_nb)
{
  int xi;
  int yi;
  dist_elt pix_elt;

  xi = (int)x;
  yi = (int)y;

  pix_elt = Tab+yi*Dimx+xi;

  pix_elt->dx     = (x-xi)*(x-xi);
  pix_elt->dy     = (y-yi)*(y-yi);
  pix_elt->src_nb = src_nb;
}

static void set_distances_priv(image_dist self,
			       bouleen compute_adjacency,
			       tab_mark_dart* mark)
{
  int i;
  dist_elt main_tab;

  /* descente */
  main_tab = Tab+Dimx;
  for(i=1;i<Dimy;i++)
    {

      traite_ligne_ascendant_priv(self,
				  main_tab,
				  main_tab-Dimx,
				  0,1);
      traite_ligne_ascendant_priv(self,
				  main_tab+1,
				  main_tab,
				  1,0);
      traite_ligne_descendant_priv(self,
				   main_tab+Dimx-2,
				   main_tab+Dimx-1,
				   -1,0,
				   faux,
				   i,
				   mark);
      main_tab += Dimx;
    }
  /* remonte */
  main_tab = Tab +Dimx*(Dimy-2);
  for(i=Dimy-1;i>0;i--)
    {

      traite_ligne_ascendant_priv(self,
				  main_tab,
				  main_tab+Dimx,
				  0,-1);
      traite_ligne_ascendant_priv(self,
				  main_tab+1,
				  main_tab,
				  1,0);
      traite_ligne_descendant_priv(self,
				   main_tab+Dimx-2,
				   main_tab+Dimx-1,
				   -1,0,
				   compute_adjacency,
				   i-1,
				   mark);
      main_tab -= Dimx;
    }

}




static void traite_ligne_ascendant_priv(image_dist self,
					dist_elt main_ligne,
					dist_elt comp_ligne,
					int incr_x,
					int incr_y)
{
  int j;
  float main_dist,comp_dist;

  for(j=incr_x;j<Dimx;j++)
    {
      /* Le point compare est a l'infini de sa source */
      if(comp_ligne->dx == -1.0)
	{
	  main_ligne++;
	  comp_ligne++;
	  continue;
	}
      /* Le point courant est a l'infini de sa source */
      if(main_ligne->dx == -1)
	{
	  main_ligne->dx     = comp_ligne->dx+incr_x;
	  main_ligne->dy     = comp_ligne->dy+incr_y;
	  main_ligne->src_nb = comp_ligne->src_nb;
	  main_ligne++;
	  comp_ligne++;
	  continue;
	}
      /* Ils sont tous les deux a distance finie : */
      /* On  calcule effectivement les distances*/
      main_dist = (main_ligne->dx*main_ligne->dx + 
		   main_ligne->dy*main_ligne->dy);
      comp_dist = (( comp_ligne->dx+incr_x)*(comp_ligne->dx + incr_x)+ 
		   (comp_ligne->dy+incr_y)*(comp_ligne->dy+incr_y));
      if(main_dist > comp_dist)
	{
	      
	  main_ligne->dx     = comp_ligne->dx+incr_x;
	  main_ligne->dy     = comp_ligne->dy+incr_y;
	  main_ligne->src_nb = comp_ligne->src_nb;
	}
      main_ligne++;
      comp_ligne++;
    }
}
static void traite_ligne_descendant_priv(image_dist self,
					 dist_elt main_ligne,
					 dist_elt comp_ligne,
					 int incr_x,
					 int incr_y,
					 bouleen compute_adjacency,
					 int ligne,
					 tab_mark_dart* mark)
{
  int j;
  float main_dist,comp_dist;
  dist_elt right,down;
  Point p,q;
  
  
  for(j=Dimx-1;j>0;j--)
    {
      /* Le point compare est a l'infini de sa source */
      if(comp_ligne->dx == -1.0)
	{
	  main_ligne--;
	  comp_ligne--;
	  continue;
	}
	  /* Le point courant est a l'infini de sa source */
      if(main_ligne->dx == -1)
	{
	  main_ligne->dx     = comp_ligne->dx+incr_x;
	  main_ligne->dy     = comp_ligne->dy+incr_y;
	  main_ligne->src_nb = comp_ligne->src_nb;
	  main_ligne--;
	  comp_ligne--;
	  continue;
	}
      /* Ils sont tous les deux a distance finie : */
      /* On  calcule effectivement les distances*/
      main_dist = (main_ligne->dx*main_ligne->dx + 
		   main_ligne->dy*main_ligne->dy);
      comp_dist = (( comp_ligne->dx+incr_x)*(comp_ligne->dx + incr_x)+ 
		   (comp_ligne->dy+incr_y)*(comp_ligne->dy+incr_y));
      if(main_dist > comp_dist)
	{
	  
	  main_ligne->dx     = comp_ligne->dx+incr_x;
	  main_ligne->dy     = comp_ligne->dy+incr_y;
	  main_ligne->src_nb = comp_ligne->src_nb;
	}
      if(compute_adjacency)
	{
	  right = comp_ligne;
	  down  = main_ligne+Dimx;
	  COORDX(p) = j-1;
	  COORDY(p) = ligne;
	  if(main_ligne->src_nb != right->src_nb)
	    {
	      update_neighbor_priv(self,mark,main_ligne->src_nb,right->src_nb);
	      tab_mark_mark_point(Frontier,&p);
	      COORDX(q) = COORDX(p)+1;
	      COORDY(q) = COORDY(p);
	      tab_mark_mark_point(Frontier,&q);
	    }
	  if(main_ligne->src_nb != down->src_nb)
	    {
	      update_neighbor_priv(self,mark,main_ligne->src_nb,down->src_nb);
	      tab_mark_mark_point(Frontier,&p);
	      COORDX(q) = COORDX(p);
	      COORDY(q) = COORDY(p)+1;
	      tab_mark_mark_point(Frontier,&q);
	    }
	}
      main_ligne--;
      comp_ligne--;
    }
}


