/* inverse_brun
 *  Copyright (C) 2000 Luc Brun
 *  email: brun@leri.univ-reims.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
static const char* sccsid = 
"@(#) dart list, Luc Brun, last revision 06/11/95";
#include <assert.h>

#include "dart_list.h"
#include "garbage.h"
#include "type_obj.h"

#define Cardinal      SELF(cardinal)
#define Nb_ligne      SELF(nb_ligne)
#define Element_list  SELF(element_list)
#define Current_ligne SELF(current_ligne)
#define Current_col   SELF(current_col)
#define Nb_of_pointer SELF(nb_of_pointer)

#define ALLOC_LIGNE    20
#define ALLOC_NB_LIGNE 50

#define NUM_LIGNE(n)   ((n)/ ALLOC_LIGNE)
#define NUM_COLONNE(n) ((n)% ALLOC_LIGNE)

static garbage line_collector;
static garbage dart_list_collector;

static void      private_insert_ligne     (dart_list,dart*);
static void      private_realloc_dart_list(dart_list);
static dart_list private_alloc_same_size  (dart_list);
/* Fonctions utilisant le garbage collector */
static void      private_free_all_dart_list   (any_ptr);
static dart*     private_alloc_line       ();
static void      private_free_line        (dart*);
static dart_list private_alloc_dart_list  ();
static void      private_free_dart_list   (dart_list);

struct dart_list
{
  dart **element_list;
  int cardinal;
  int nb_ligne;
  int current_ligne;
  int current_col;
  int nb_of_pointer;
};
void ALLOC_dart_list()
{
  line_collector = FAIRE_garbage();
  dart_list_collector = FAIRE_garbage();
}
void DEALLOC_dart_list()
{
  DEFAIRE_garbage(line_collector,free);
  DEFAIRE_garbage(dart_list_collector,private_free_all_dart_list);
}
dart_list FAIRE_dart_list()
{
  dart_list self;

  self = private_alloc_dart_list();

  *Element_list= private_alloc_line();
  Cardinal = 0;
  Nb_ligne = 1;
  Current_ligne = Current_col =0;
  
  return self;
}

/*---------------------------------------------------------------------------*/

void DEFAIRE_dart_list(dart_list self)
{  
  dart **tmp;
  int i;

  assert(self);

  tmp = Element_list;

  for(i=0;i<Nb_ligne;i++)
    {
      private_free_line(*tmp);
      *tmp = NULL;
      tmp++;
    }
  private_free_dart_list(self);
}

/*---------------------------------------------------------------------------*/
void dart_list_reinitialize(dart_list self)
{
  Cardinal = 0;
  Current_ligne = 0;
  Current_col =0;
}

/*---------------------------------------------------------------------------*/
int dart_list_give_cardinal(dart_list self)
{
  assert(self);
  return Cardinal;
}
/*---------------------------------------------------------------------------*/
dart_list dart_list_copy(dart_list self)
{
  dart_list copy;
  dart* tmp,*tmp_copy;
  int nb_ligne_pleine,reste;
  int i,j;

  assert(self);

  copy = private_alloc_same_size(self);

  nb_ligne_pleine     = NUM_LIGNE(Cardinal);

  copy->cardinal      = Cardinal;
  copy->nb_ligne      = nb_ligne_pleine+1;
  copy->current_ligne = Current_ligne;
  copy->current_col   = Current_col;




  /* On parcourt les elts de list pour les inserer dans self */
  for(i=0;i<nb_ligne_pleine;i++)
    {   
      tmp = Element_list[i];
      tmp_copy = copy->element_list[i];
      for(j=0;j<ALLOC_LIGNE;j++)
	{
	  *tmp_copy = *tmp;
	  tmp_copy++;
	  tmp++;
	}
    }  
/* On parcourt la partie remplie du dernier tableau */

  tmp      = Element_list[nb_ligne_pleine];
  tmp_copy = copy->element_list[nb_ligne_pleine];

  reste    = NUM_COLONNE(Cardinal);

  for(j=0;j<reste;j++)
    {
      *tmp_copy = *tmp;
      tmp_copy++;
      tmp++;
    }
  
  return copy;
}
  /*-------------------------------------------------------------------*/
 /* Concatenation de deux listes, mise du resultat dans une troisieme */
/*-------------------------------------------------------------------*/

dart_list dart_list_concat(dart_list lfirst, dart_list llast)
{
  dart_list concat;

  assert(lfirst);
  assert(llast);

  concat = dart_list_copy(lfirst);
  dart_list_insert_list(concat,llast);
  
  return concat;
}

void dart_list_first(dart_list self)
{
  assert(self);

  if(Cardinal == 0)
    return;

  Current_ligne = 0;
  Current_col   = 0;
}

void dart_list_last(dart_list self)
{
  int nb_ligne_pleine;
  int reste;

  assert(self);

  if(Cardinal == 0)
    return;

  nb_ligne_pleine = NUM_LIGNE(Cardinal);
  reste           = NUM_COLONNE(Cardinal);

  if(reste == 0)
    {
      Current_ligne = nb_ligne_pleine-1;
      Current_col   = ALLOC_LIGNE-1;

      assert(Current_ligne >=0);

      return;
     }      

  Current_ligne = nb_ligne_pleine; 
  Current_col   = reste-1;
}

dart dart_list_give_current(dart_list self)
{
  assert(self);

  return Element_list[Current_ligne][Current_col];
}
 /* Insere un element dans la liste */
/* Remarque le meme element peut etre insere plusieur fois */
void dart_list_insert(dart_list self, dart element)
{
  int ligne,col;

  assert(self);

  if( Cardinal == Nb_ligne*ALLOC_LIGNE)
    private_realloc_dart_list(self);
  
  ligne = NUM_LIGNE(Cardinal);
  col   = NUM_COLONNE(Cardinal);
  
  Element_list[ligne][col] = element;
  Cardinal++;
}
  /*----------------------------------------*/
 /* Insertion des elts de  list dans self  */
/*----------------------------------------*/
void dart_list_insert_list(dart_list self,dart_list list)
{
  int i,j;
  int nb_ligne_pleine;
  int reste;
  dart* tmp;

  assert(self);
  assert(list);

  nb_ligne_pleine = NUM_LIGNE(list->cardinal);
  reste           = NUM_COLONNE(list->cardinal);

  /* On parcourt les elts de list pour les inserer dans self */
  for(i=0;i<nb_ligne_pleine;i++)
    {
      tmp = list->element_list[i];
      for(j=0;j<ALLOC_LIGNE;j++)
	{
	  dart_list_insert(self,*tmp);
	  tmp++;
	}
    }
  /* On parcourt la partie remplie du dernier tableau */
  tmp = list->element_list[nb_ligne_pleine];
  for(j=0;j<reste;j++)
    {
      dart_list_insert(self,*tmp);
      tmp++;
    }
}
/*******************************************************************/
/* meme fonctionalitees que insert mais detruit le second argument*/
/*****************************************************************/
void dart_list_merge_list(dart_list self,dart_list list)
{
  int i;
  int nb_ligne_pleine;
  int reste;
  dart* tmp,*last_ligne;
  
  /* On parcourt la partie remplie du dernier tableau et on l'insere */
  /* dans self */
  assert(self);
  assert(list);
  
  if (dart_list_is_empty(list))
    {
      DEFAIRE_dart_list(list);
      return;
    }

  nb_ligne_pleine =  NUM_LIGNE(list->cardinal);
  reste           =  NUM_COLONNE(list->cardinal);

  tmp = list->element_list[nb_ligne_pleine];
  for(i=0;i<reste;i++)
    {
      dart_list_insert(self,*tmp);
      tmp++;
    }
  
  /* list est a present constitue uniquement de lignes pleines. */
  /* On oublie temporairement le dernier tableau de self sauve ds */
  /* last_ligne*/

  for(i=0;i<nb_ligne_pleine;i++)
    {
      private_insert_ligne(self,list->element_list[i]);
      Cardinal += ALLOC_LIGNE;
    }

  /* Si liste contenait des tableaux vides on les effaces*/
  for(i=nb_ligne_pleine;i<list->nb_ligne;i++)
    private_free_line(list->element_list[i]);
  
  private_free_dart_list(list);
}

void dart_list_remove_list(dart_list self,dart_list lremove)
{
  int i,j;
  int nb_ligne_pleine;
  int reste;
  dart* tmp;

  assert(self);

  nb_ligne_pleine = NUM_LIGNE(lremove->cardinal);
  reste           = NUM_COLONNE(lremove->cardinal);

  /* On parcourt les elts de lremove pour les enlever (eventuellement) de self */
  for(i=0;i<nb_ligne_pleine;i++)
    {
      tmp = lremove->element_list[i];
      for(j=0;j<ALLOC_LIGNE;j++)
	{
	  dart_list_remove_elt(self,*tmp);
	  tmp++;
	}
    }
  /* On parcourt la partie remplie du dernier tableau */
  tmp = lremove->element_list[nb_ligne_pleine];
  for(j=0;j<reste;j++)
    {
      dart_list_remove_elt(self,*tmp);
      tmp++;
    }
}
/* Calcule l'intersection de deux liste */
/* Rem1 : Un element apparaissant 2 fois dans une list et une fois */
/* dans self  apparaitra 2 fois dans la liste resultante. */
/* Rem 2 : Rarement vu un machin aussi couteux */

dart_list dart_list_intersection(dart_list self, dart_list list)
{
  
  dart_list  intersect = FAIRE_dart_list();
  dart* tmp;
  int nb_ligne_pleine;
  int reste;
  int i,j;

  assert(self);
  
  nb_ligne_pleine = NUM_LIGNE(Cardinal);
  reste          =  NUM_COLONNE(Cardinal);

  /* On cherche les elts d'intersection dans les tab pleins */
  for(i=0;i<nb_ligne_pleine;i++)
    {
      tmp = Element_list[i];
      for(j=0;j<ALLOC_LIGNE;j++)
	{
	  if( dart_list_is_in(list,*tmp) )
	    dart_list_insert(intersect,*tmp);
	  tmp++;
	}
    }
  /* On les cherche dans le dernier tab */
  tmp = Element_list[nb_ligne_pleine];
  for(j=0;j<reste;j++)
    {
      if(dart_list_is_in(list,*tmp))
	dart_list_insert(intersect,*tmp);
      tmp++;
    }
  /* Si l'intersection est vide on renvoie null */
  if(intersect->cardinal==0)
    {
      DEFAIRE_dart_list(intersect);
      return (dart_list)NULL;
    }

  return intersect;
}
  
bouleen dart_list_next(dart_list self)
{
  int reste,ligne_remplie,ligne,col; 
  bouleen depassement_ligne,depassement_col;

  assert(self);

  ligne_remplie = NUM_LIGNE(Cardinal);
  reste         = NUM_COLONNE(Cardinal);

  ligne         = Current_ligne;
  col           = Current_col;

  col++;

  /* On etait au bout d'une ligne */
  if(col == ALLOC_LIGNE)
    {
      col =0;
      ligne++;
    }

  depassement_ligne = (ligne > ligne_remplie);
  depassement_col   = (ligne == ligne_remplie)&&(col >= reste);
  
  if(depassement_ligne||depassement_col)
    return faux;
  
  Current_ligne = ligne;
  Current_col = col;
  
  return vrai;
}

bouleen dart_list_previous(dart_list self)
{
  assert(self);

  if( (Current_col == 0) && (Current_ligne ==0) ) return faux;
  
  Current_col--;
  if(Current_col<0)
    {
      Current_col = ALLOC_LIGNE-1;
      Current_ligne --;
    }

  return vrai;
}
bouleen dart_list_is_empty(dart_list self)
{
  assert(self);

  return (Cardinal == 0);
}
bouleen dart_list_is_in(dart_list self, dart b)
{
  dart* tmp;
  int reste,ligne_remplie;
  int i,j;

  assert(self);
  
  ligne_remplie = NUM_LIGNE(Cardinal);
  reste         = NUM_COLONNE(Cardinal);

  i=0;
  
  while (i< ligne_remplie) 
    {
      j=0;
      tmp = Element_list[i];
      while ( j<ALLOC_LIGNE) 
	{
	  if(*tmp == b)
	    return vrai;

	  tmp++;
	  j++;
	}
      i++;
    }
  j=0;
  tmp = Element_list[ligne_remplie];
  while (j<reste) 
    {
      if(*tmp == b)
	return vrai;

      tmp++;
      j++;
    }

  return faux;

}

/* Supprime un element de la liste. */
/* Rem : Si celui ci se trouve en plusieurs exemplaires dans la liste */
/* une seule occurence sera supprime */
/* Pour supprimer toutes les occurences d'une liste il suffit de faire */
/* while(dart_list_remove_elt(..));*/

bouleen dart_list_remove_elt(dart_list self, dart b)
{
  dart* tmp;
  int reste,ligne_remplie;
  dart dernier;
  int i,j;
  
  assert(self);
  
  if(Cardinal == 0)
    return faux;
  
  ligne_remplie = NUM_LIGNE(Cardinal);
  reste         = NUM_COLONNE(Cardinal);
  if(reste==0)
    dernier = Element_list[ligne_remplie-1][ALLOC_LIGNE-1];
  else
    dernier = Element_list[ligne_remplie][reste-1];
  i=0;

  while (i< ligne_remplie)
    {
      j=0;
      tmp = Element_list[i];
      while (j<ALLOC_LIGNE) 
	{
	  if(*tmp == b)
	    {
	      Cardinal--;
	      *tmp = dernier;
	      return vrai;
	    }
	  tmp++;
	  j++;
	}
      i++;
    }
  j=0;
  tmp = Element_list[ligne_remplie];
  while (j<reste) 
    {
      if(*tmp == b)
	{
	  Cardinal--;
	  *tmp = dernier;
	  return vrai;
	}
      tmp++;
      j++;
    }

  return faux;
  
}

void dart_list_print(dart_list self)
{
  dart b;

  if( dart_list_is_empty(self) )
    {
      printf("( )\n");
      return;
    }
  dart_list_first(self);
  b = dart_list_give_current(self);
  printf("\n(%d",b);

  while(dart_list_next(self))
    {
      b = dart_list_give_current(self);
      printf(", %d",b);
    }
  
  printf(")\n");
}
/* On alloue une dart list de la meme taille que self */
static dart_list private_alloc_same_size(dart_list self)
{
  dart_list list2;
  int i;
  int ligne_a_allouer;
  int nb_tab;

  ligne_a_allouer = NUM_LIGNE(Cardinal)+1;
  nb_tab = ligne_a_allouer/ALLOC_NB_LIGNE+1;

  list2 = private_alloc_dart_list();

  if (ligne_a_allouer > list2->nb_of_pointer)
    {
      list2->element_list = (dart **)realloc(list2->element_list,
					     nb_tab*ALLOC_NB_LIGNE*
					     sizeof(dart*));

      list2->nb_of_pointer = nb_tab*ALLOC_NB_LIGNE;
    }
  for(i=0;i<ligne_a_allouer;i++)
    list2->element_list[i] = private_alloc_line();
  
  return list2;
}

static void private_realloc_dart_list(dart_list self)
{


  /* On regarde si on na pas depasse le nb max de ligne alloue */
  if (Nb_ligne == Nb_of_pointer)
    {
      Element_list = (dart**)realloc(Element_list,
				     (Nb_ligne+ALLOC_NB_LIGNE)*
				     sizeof(dart*));
      assert(Element_list);
      Nb_of_pointer += ALLOC_NB_LIGNE;
    }
  
  Element_list[Nb_ligne] = private_alloc_line();
  Nb_ligne++;
}

static void private_insert_ligne(dart_list self,dart* ligne)
{
  int last_line;

  assert(self);

  /* On regarde si on na pas depasse le nb max de ligne alloue */
  if(Nb_ligne == Nb_of_pointer)
    {
      Element_list = (dart**)realloc(Element_list,
				     (Nb_ligne+ALLOC_NB_LIGNE)*
				     sizeof(dart*));
      assert(Element_list);
      Nb_of_pointer += ALLOC_NB_LIGNE;
    }

  last_line = NUM_LIGNE(Cardinal);
  if((NUM_COLONNE(Cardinal)==0)&&(Cardinal > ALLOC_LIGNE))
    last_line--;

  Element_list[Nb_ligne]    = Element_list[last_line+1];
  Element_list[last_line+1] = Element_list[last_line];
  Element_list[last_line]   = ligne;
  
  Nb_ligne++;
}
/* Utilise pour vider le garbage en fin de programe */
static void private_free_all_dart_list(any_ptr obj)
{
  dart_list self = obj;
  dart **tmp;
  int i;
  
  assert(self);
  
  tmp = Element_list;
  
  for(i=0;i<Nb_ligne;i++)
    {
      free(*tmp);
      tmp++;
    }
  free(self);
}
static dart* private_alloc_line()
{
  dart* line;

  if (garbage_is_empty(line_collector))
    return  (dart *) malloc(ALLOC_LIGNE*sizeof(dart));

  return garbage_give_object(line_collector);
}
static void private_free_line(dart* line)
{
  garbage_insert_object(line_collector,line);
}

static dart_list private_alloc_dart_list()
{
  dart_list self;

  if (garbage_is_empty(dart_list_collector))
    {
      self =  (dart_list)malloc(sizeof(struct dart_list));
      Element_list  = (dart **) malloc(ALLOC_NB_LIGNE*sizeof(dart*));      
      Nb_of_pointer = ALLOC_NB_LIGNE;
      return self;
    }

  return  garbage_give_object(dart_list_collector);
}
static void private_free_dart_list(dart_list self)
{
  garbage_insert_object(dart_list_collector,self);
}
