/* inverse_brun
 *  Copyright (C) 2000 Luc Brun
 *  email: brun@leri.univ-reims.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
static const char* sccsid = 
"@(#) generic list, Luc Brun, last revision 14/11/95";

#include <stdio.h>
#include <assert.h>

#include "list_gen.h"
#include "objet.h"

static list_gen line_collector = NULL_LIST_GEN;
static list_gen list_gen_collector = NULL_LIST_GEN;


#define Tab_list       SELF(tab_list)
#define Current_ligne  SELF(current_ligne)
#define Current_column SELF(current_column)
#define Cardinal       SELF(cardinal)
#define Nb_ligne       SELF(nb_ligne)
#define Nb_of_pointer  SELF(nb_of_pointer)

#define ALLOC_LIGNE    20
#define ALLOC_NB_LIGNE 50

#define NUM_LIGNE(n) ((n)/(ALLOC_LIGNE))
#define NUM_COLONNE(n) ((n)%ALLOC_LIGNE)

static void     realloc_list_gen_priv  (list_gen);
static void     free_elts_priv         (list_gen,void (*del)(void*)); 
static list_gen alloc_list_gen_priv    ();
static void     dealloc_list_gen_priv  (list_gen,void (*del)(void*));
static any_ptr* alloc_line_priv        ();
static void     dealloc_line_priv      (any_ptr*);
static void     free_list_gen_priv     (list_gen,void (*del)(void*));
static void     free_tab_list_priv     (void*);
static list_gen alloc_new_list_gen_priv();

struct list_gen
{    
  any_ptr     **tab_list;
  int      current_ligne;
  int      current_column;
  int      cardinal;
  int      nb_ligne;
  int      nb_of_pointer;
};
void ALLOC_list_gen()
{
  line_collector     = alloc_new_list_gen_priv();
  list_gen_collector = alloc_new_list_gen_priv();
}
void DEALLOC_list_gen()
{
  free_list_gen_priv(line_collector,free);
  free_list_gen_priv(list_gen_collector,free_tab_list_priv);
}
list_gen FAIRE_list_gen()
{
  list_gen self;

  self = alloc_list_gen_priv();

  return self;
}

/*---------------------------------------------------------------------------*/
/* On libere les differents tableaux */ 
/*---------------------------------------------------------------------------*/
void DEFAIRE_list_gen(list_gen self,void (*del)(void*))
{
  dealloc_list_gen_priv(self,del);
}

/*---------------------------------------------------------------------------*/
void list_gen_reinitialize(list_gen self)
{
  int i;
  
  for(i=1;i<Nb_ligne;i++)
    dealloc_line_priv(Tab_list[i]);
  

  Cardinal = 0;
  Nb_ligne = 1;
  Current_ligne  = 0;
  Current_column = 0;
}
/*---------------------------------------------------------------------------*/
void list_gen_insert(list_gen self, any_ptr elt)
{
  int ligne,column;


  ligne  = NUM_LIGNE  (Cardinal);
  column = NUM_COLONNE(Cardinal);
  
  /* On tombe sur un ligne non alloue */
  if (ligne >= Nb_ligne)
    {
      if(self == line_collector)
	{
	  if(Nb_ligne == Nb_of_pointer)
	    {
	      Tab_list = (any_ptr**)realloc(Tab_list,
					    (Nb_ligne+ALLOC_NB_LIGNE)*
					    sizeof(any_ptr*));
	      assert(Tab_list);
	      Nb_of_pointer += ALLOC_NB_LIGNE;
	    }
	  Tab_list[Nb_ligne] = elt;
	  Nb_ligne++;
	  return;
	}
      realloc_list_gen_priv(self);
    }
  Tab_list[ligne][column] = elt;
  Cardinal++;
  
}
/*---------------------------------------------------------------------------*/
void list_gen_copy(list_gen self,list_gen copy)
{
  any_ptr obj;

  assert(self);
  assert(copy);

  list_gen_reinitialize(copy);

  if (list_gen_is_empty(self))
    return;
  
  list_gen_first(self);
  do
    {
      obj = list_gen_give_current(self);
      list_gen_insert(copy,obj);
    }
  while (list_gen_next(self));
}
/*---------------------------------------------------------------------------*/

any_ptr list_gen_give_current(list_gen self)
{
  return Tab_list[Current_ligne][Current_column];
}
/*---------------------------------------------------------------------------*/
void list_gen_first(list_gen self)
{
  Current_ligne = 0;
  Current_column = 0;
}
/*---------------------------------------------------------------------------*/
void list_gen_last(list_gen self)
{
  int line,column;

  line = NUM_LIGNE(Cardinal);
  column = NUM_COLONNE(Cardinal)-1;

  if(column == -1)
    {
      column = ALLOC_LIGNE-1;
      line -=1;
    }
  
  Current_ligne = line;
  Current_column = column;
}
  
/*---------------------------------------------------------------------------*/
bouleen list_gen_next(list_gen self)
{
  int ligne = Current_ligne;
  int column = Current_column+1;

  /* On se preocupe du saut d'un talbeau a l'autre */
  if(column == ALLOC_LIGNE)
    {
      ligne++;
      column =0;
    }
  /* On regarde si la case envisage est alloue */
  if((ligne >= Nb_ligne) || (ligne*ALLOC_LIGNE+column >= Cardinal) )
    return faux;

  Current_ligne = ligne;
  Current_column = column;
  return vrai;
}
bouleen list_gen_previous(list_gen self)
{
  int ligne = Current_ligne;
  int column = Current_column-1;

  /* On se preocupe du saut d'un talbeau a l'autre */
  if(column <0)
    {
      ligne--;
      column =ALLOC_LIGNE-1;
    }
  /* On regarde si on a pas depasse le debut */
  if( ligne < 0 )
    return faux;

  Current_ligne = ligne;
  Current_column = column;

  return vrai;
}

/*---------------------------------------------------------------------------*/
/* On efface l'element courant en mettant le dernier a la place de
l'element courant et en effacant le dernier*/
/* ATTENTION : On efface pas l'objet de la memoire. Cela doit etre */
/* fait separement */
/*---------------------------------------------------------------------------*/
void list_gen_remove_current(list_gen self) 
{ 

  int ligne,column;
  Cardinal--;

  if (Cardinal == 0)
    {
      assert(Current_ligne == 0);
      assert(Current_column == 0);
      Tab_list[0][0] = NULL;
      return;
    }

  ligne  = NUM_LIGNE(Cardinal);
  column = NUM_COLONNE(Cardinal);
  
  Tab_list[Current_ligne][Current_column] =  Tab_list[ligne][column];
  assert(Tab_list[ligne][column]);
  Tab_list[ligne][column] = NULL;
  
  if (self == line_collector)
    return;

  if (Cardinal%ALLOC_LIGNE == 0)
    {
      Nb_ligne--;
      dealloc_line_priv(Tab_list[Nb_ligne]);
      Tab_list[Nb_ligne] = NULL;
    }
}
bouleen list_gen_remove_object(list_gen self,
			       PFequal equal,
			       any_ptr obj)

{
  any_ptr current_obj;

  assert(self);
  assert(equal);
  assert(obj);
  
  if(Cardinal == 0)
    return faux;

  list_gen_first(self);
  do
    {
      current_obj = list_gen_give_current(self);
      if(equal(current_obj,obj))
	{
	  list_gen_remove_current(self);
	  return vrai;
	}
    }
  while(list_gen_next(self));

  return faux;
}
/*---------------------------------------------------------------------------*/
int list_gen_give_cardinal(list_gen self)
{
  assert(self);

  return Cardinal;
}
bouleen list_gen_is_empty(list_gen self)
{
  assert(self);

  return (Cardinal == 0);
}
/*---------------------------------------------------------------------------*/
static void realloc_list_gen_priv(list_gen self)
{
  /* On regarde si on na pas depasse le nb max de ligne alloue */
  if(Nb_ligne == Nb_of_pointer)
    {
      Tab_list = (any_ptr**)realloc(Tab_list,
				   (Nb_ligne+ALLOC_NB_LIGNE)*
				   sizeof(any_ptr*));
      assert(Tab_list);
      Nb_of_pointer += ALLOC_NB_LIGNE;
    }
  Tab_list[Nb_ligne] = alloc_line_priv();
  assert(Tab_list[Nb_ligne]);
  Nb_ligne++;
}
/* On efface les elements composant la liste a l'aide de la fonction */
/* del fournie par l'utilisateur */
static void free_elts_priv(list_gen self,void (*del)(void*))
{
  int reste;
  int i,j;
  any_ptr *tmp;

  if(Cardinal == 0)
    return;

  for(i=0;i<Nb_ligne-1;i++)
    {
      tmp = Tab_list[i];
      for(j=0;j<ALLOC_LIGNE;j++)
	{
	  del(*tmp);
	  *tmp = NULL;
	  tmp++;
	}
    }
  reste = Cardinal - (Nb_ligne-1)*ALLOC_LIGNE;
  tmp = Tab_list[Nb_ligne-1];
  for(i=0;i<reste;i++)
    {
      del(*tmp);
      *tmp = NULL;
      tmp++;
    }
}
static list_gen alloc_list_gen_priv()
{
 list_gen self;

 if(list_gen_is_empty(list_gen_collector))
   {
     self = alloc_new_list_gen_priv();
     return self;
   }

 list_gen_first(list_gen_collector);
 self = list_gen_give_current(list_gen_collector);
 list_gen_remove_current(list_gen_collector);
 
 return self;
}    
static list_gen alloc_new_list_gen_priv()
{
  list_gen self;
  
  self = (list_gen)malloc(sizeof(struct list_gen));
  assert(self);
  Tab_list = (any_ptr **) calloc(ALLOC_NB_LIGNE,
				 sizeof(any_ptr*));
  *Tab_list = alloc_line_priv();
  
  Cardinal = 0;
  Nb_ligne = 1;
  Nb_of_pointer = ALLOC_NB_LIGNE;
  assert(Tab_list);assert(*Tab_list);

  return self;
}   

     
static void dealloc_list_gen_priv(list_gen self,void (*del)(void*))
{
  any_ptr **tmp;
  int i;
  
  free_elts_priv(self,del);

  tmp  = Tab_list+1;
  for(i=1;i<Nb_ligne;i++)
    {
      dealloc_line_priv(*tmp);
      tmp++;
    }

  Cardinal = 0;
  Nb_ligne = 1;
  Current_ligne  = 0;
  Current_column = 0;

  list_gen_insert(list_gen_collector,self);
}
static any_ptr* alloc_line_priv()
{
  any_ptr* line;
  
  if(line_collector == NULL_LIST_GEN)
    {      
      line = (any_ptr*) calloc(ALLOC_LIGNE,
			       sizeof(any_ptr));
      assert(line);
      return line;
    }

  if(list_gen_is_empty(line_collector))
    {
      line = (any_ptr*) calloc(ALLOC_LIGNE,
			       sizeof(any_ptr));
      assert(line);
      return line;
    }
  assert(line_collector->tab_list[0][0]);
  list_gen_first(line_collector);
  line = (any_ptr*) list_gen_give_current(line_collector);
  list_gen_remove_current(line_collector);
  
  assert(line);
  return line;
}
static void dealloc_line_priv(any_ptr* line)
{
  assert(line);
  list_gen_insert(line_collector,line);
}
static void free_list_gen_priv(list_gen self,void (*del)(void*))
{
  int i;
  
  free_elts_priv(self,del);

  for(i=0;i<Nb_ligne;i++)
    free(Tab_list[i]);
  
  free(Tab_list);
  free(self);
}
static void free_tab_list_priv(void* list)
{
  list_gen self = list;
  int i;
  assert(self);

  for(i=0;i<Nb_ligne;i++)
    free(Tab_list[i]);
  
  free(Tab_list);
  free(self);
}







