/* inverse_brun
 *  Copyright (C) 2000 Luc Brun
 *  email: brun@leri.univ-reims.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include <stdio.h>
#include <assert.h>

#include "tab_mark.h"
#include "point.h"
#include "objet.h"

#define Bitm           SELF (bitm)
#define DimX           SELF (dimX)
#define DimY           SELF (dimY)
#define X_max          SELF (x_max)
#define Y_max          SELF (y_max)
#define Limit          SELF (limit)
#define Freebit        SELF (freebit)


static  any  pixel_mask[4]     = {0x03,0x0c,0x30,0xc0};
static  any  pixel_mask_one[4] = {0x01,0x04,0x10,0x40};
static  any  pixel_mask_two[4] = {0x02,0x08,0x20,0x80};

static bouleen private_go_point(tab_mark,point,long*,int*);

/****************************************************************************/
/* Les pixels sont codes sur 2 bits. Donc un octet(un any) represente */
/* 4 pixels */
/****************************************************************************/
struct tab_mark 
{
  any    *freebit;
  any    *bitm;
  any    *limit;
  long   dimX;
  long   dimY;
  long   x_max;
  long   y_max;
};

tab_mark FAIRE_tab_mark(long dimy,long dimx)
{
  long  integer,rest;
  
  INSTANCIER(tab_mark);
  
  /* On ajoute 1 a dimx et dimy pour pouvoir eventuellement travailler */
  /* dans P1/2 */
  DimX  = dimx+1;
  DimY  = dimy+1;
  
  X_max = DimX - 1;
  Y_max = DimY - 1;
  
  
  integer  = DimX * DimY;
  rest     = integer&0x3; /* reste modulo 4 */
  integer =  integer>>2; /* division entiere par 4 */
  
  Bitm  = (any *) calloc(integer + rest,1);
  Limit    = Bitm + integer + rest;
  
  return self;
}

/* DimX and DimY code the coordinates */

/*---------------------------------------------------------------------------*/

void DEFAIRE_tab_mark(tab_mark self)
{
  free (Bitm);
  free(self);
}
/*---------------------------------------------------------------------------*/
int tab_mark_give_hauteur(tab_mark self)
{
  return (DimY-1);
}
/*---------------------------------------------------------------------------*/
int tab_mark_give_largeur(tab_mark self)
{
  return (DimX-1);
}

/*---------------------------------------------------------------------------*/
/* Efface toute marque du tableau */
/*---------------------------------------------------------------------------*/
void tab_mark_clear(tab_mark self)
{
  any *pt1;
  
  pt1 = Bitm;
  while (pt1 < Limit) *pt1++ = 0;
}

/*---------------------------------------------------------------------------*/
/* Positionne la marque 1 au point p */
/*---------------------------------------------------------------------------*/
void tab_mark_mark_point(tab_mark self,point p)
{
  long move;
  int offset;
  any *pixel_set;
  
  assert(self);
  assert(p);
  assert( private_go_point(self,p,&move,&offset) );

  pixel_set = Bitm + move;
  
  *pixel_set |= pixel_mask_one[offset];
}

/*---------------------------------------------------------------------------*/
/* Positionne la marque 2 au point p */
/*---------------------------------------------------------------------------*/
void tab_mark_mark_point_two(tab_mark self,point p)
{
  long move;
  int offset;
  any *pixel_set;
  
  assert(self);
  assert(p);
  assert( private_go_point(self,p,&move,&offset) );

  pixel_set = Bitm + move;
      
  *pixel_set |= pixel_mask_two[offset];

}
/*---------------------------------------------------------------------------*/
/* Enleve toute marque au point p */
/*---------------------------------------------------------------------------*/
void tab_mark_unmark_point_all(tab_mark self,point p)
{
  long move;
  int offset;
  any *pixel_set;
  
  assert(self);
  assert(p);
  assert( private_go_point(self,p,&move,&offset) );

  pixel_set = Bitm + move;
  
  *pixel_set &= ~pixel_mask[offset];

}
void tab_mark_unmark_point(tab_mark self,point p)
{
  long move;
  int offset;
  any *pixel_set;
  
  assert(self);
  assert(p);
  assert( private_go_point(self,p,&move,&offset) );

  pixel_set = Bitm + move;
  
  *pixel_set &= ~pixel_mask_one[offset];

}
void tab_mark_unmark_point_two(tab_mark self,point p)
{
  long move;
  int offset;
  any *pixel_set;
  
  assert(self);
  assert(p);
  assert( private_go_point(self,p,&move,&offset) );

  pixel_set = Bitm + move;
  
  *pixel_set &= ~pixel_mask_two[offset];

}
  /*-------------------------------------------------*/
 /* Indique si la marque 1 et positionne au point p */
/*-------------------------------------------------*/
bouleen  tab_mark_is_marked(tab_mark self,point p)
{
  long move;
  int offset;
  any *pixel_set;
  bouleen marked;
  
  assert(self);
  assert(p);

  assert( private_go_point(self,p,&move,&offset) );


  pixel_set = Bitm + move;
  
  marked = (*pixel_set)&pixel_mask_one[offset];

  return (marked !=0);
}

  /*-------------------------------------------------*/
 /* Indique si la marque 2 et positionne au point p */
/*-------------------------------------------------*/
bouleen  tab_mark_is_marked_two(tab_mark self,point p)
{
  long move;
  int offset;
  any *pixel_set;
  bouleen marked;

  assert(self);
  assert(p);

  assert( private_go_point(self,p,&move,&offset) );


  pixel_set = Bitm + move;
  
  marked = *pixel_set&pixel_mask_two[offset];
  
  return (marked !=0);
}

/*---------------------------------------------------------------------------*/
/* Calcule le deplacement et l'offset necessaires pour atteindre le */
/* point */
/*---------------------------------------------------------------------------*/
static bouleen private_go_point(tab_mark self,point p,long* move,int* offset)
{

  long dep;
  Point p_bitmap;

  if( NULL_POINT(p) )
    return faux;

  if(INFINITE_POINT(*p))
    return faux;

  COORDX(p_bitmap) = COORDX(*p)+1;
  COORDY(p_bitmap) = COORDY(*p)+1;

  if ( (COORDX(p_bitmap) > X_max) || (COORDY(p_bitmap) > Y_max) )
    return faux;

  dep     = DimX*COORDY(p_bitmap)  + COORDX(p_bitmap);
  *offset = dep&0x3; /* reste modulo 4 */
  *move   = dep>>2; /* division entiere par 4 */

  return vrai;
}






