/* inverse_brun
 *  Copyright (C) 2000 Luc Brun
 *  email: brun@leri.univ-reims.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
static const char* sccsid = 
"@(#) tab_mark_dart, used to mark, unmark a set of arc.Luc Brun, last revision 04/12/95";

#include <assert.h>
#include <stdlib.h>
#include "tab_mark_dart.h"
#include "garbage.h"
#include "macros.h"

#include "objet.h"

#define Nb_line       SELF(nb_line)
#define Nb_of_pointer SELF(nb_of_pointer)
#define Element_tab   SELF(element_tab)


#define ALLOC_LINE 250
#define NB_MARK_PER_BYTE 4
#define NB_MARK_PER_LINE 1000 /* NB_MARK_PER_BYTE*ALLOC_LINE*/
#define ALLOC_NB_LINE 10

#define NUM_LINE(n)   ((n)  / NB_MARK_PER_LINE)
#define NUM_COLUMN(n) (((n) % NB_MARK_PER_LINE)/NB_MARK_PER_BYTE)
#define NUM_MASK(n)   (((n) % NB_MARK_PER_LINE)%NB_MARK_PER_BYTE)

typedef unsigned char four_mark; 

#define NULL_FOUR_MARK (four_mark)0;

static void  realloc_tab_mark_dart_priv(tab_mark_dart);
/* Fonctions liees au garbage */
static four_mark*    alloc_line_priv             ();
static tab_mark_dart alloc_tab_mark_dart_priv    ();
static void          dealloc_line_priv           (four_mark*);
static void          dealloc_tab_mark_dart_priv  (tab_mark_dart);
static void          free_tab_mark_dart_priv     (void*);

static garbage line_collector     = NULL_GARBAGE;
static garbage tab_mark_dart_collector = NULL_GARBAGE;

static unsigned char give_put_mask[2][NB_MARK_PER_BYTE] =
{
  {0x01,0x04,0x10,0x40},
  {0x02,0x08,0x20,0x80}
};

struct tab_mark_dart
{
  four_mark **element_tab;
  int nb_line;
  int nb_of_pointer;
};

void ALLOC_tab_mark_dart()
{
  if(line_collector == NULL_GARBAGE)
    line_collector = FAIRE_garbage();
  if(tab_mark_dart_collector == NULL_GARBAGE)
    tab_mark_dart_collector = FAIRE_garbage();
}
void DEALLOC_tab_mark_dart()
{
  if(line_collector != NULL_GARBAGE)
    {
      DEFAIRE_garbage(line_collector,free);
      line_collector = NULL_GARBAGE;
    }
  if(tab_mark_dart_collector != NULL_GARBAGE)
    {
      DEFAIRE_garbage(tab_mark_dart_collector,free_tab_mark_dart_priv);
      tab_mark_dart_collector = NULL_GARBAGE;
    }
}
tab_mark_dart FAIRE_tab_mark_dart()
{
  tab_mark_dart self;
  
  self           = alloc_tab_mark_dart_priv();
  *Element_tab   = alloc_line_priv();
  Nb_line        = 1;

  return self;
}

/*---------------------------------------------------------------------------*/

void DEFAIRE_tab_mark_dart(tab_mark_dart self)
{  
  int i;

  assert(self);

  for(i=0;i<Nb_line;i++)
    dealloc_line_priv(Element_tab[i]);

  dealloc_tab_mark_dart_priv(self);

}

void tab_mark_dart_reinitialize(tab_mark_dart self)
{
  int i;
  
  assert(self);
  
  for(i=1;i<Nb_line;i++)
    dealloc_line_priv(Element_tab[i]);
  
  for(i=0;i<ALLOC_LINE;i++)
    Element_tab[0][i] = NULL_FOUR_MARK;

  Nb_line = 1;
}
/*---------------------------------------------------------------------------*/
/* Indique si le brin b est marque */
/*---------------------------------------------------------------------------*/
bouleen  tab_mark_dart_is_marked(tab_mark_dart self,dart b)
{
  int line,column,mask;
  four_mark mark;
  dart b_abs;

  assert(self);
  
  b_abs = ABS(b);

  line   = NUM_LINE(b_abs);
  column = NUM_COLUMN(b_abs);
  mask   = NUM_MASK(b_abs); 
  
  /* ligne non encore alloue */
  if(line >= Nb_line)
    return faux;

  mark = Element_tab[line][column]&give_put_mask[b<0][mask];

  /* On le ramene sur les 2 premiers bits de l'octet */
  
  return (mark != 0);
}
/*---------------------------------------------------------------------------*/
/* met value a l'indice ind du tableau virtuel */
/*---------------------------------------------------------------------------*/
void tab_mark_dart_mark_dart(tab_mark_dart self,dart b)
{
  int line,column,mask;
  dart b_abs;

  assert(self);
  
  b_abs = ABS(b);

  line   = NUM_LINE(b_abs);
  column = NUM_COLUMN(b_abs);
  mask   = NUM_MASK(b_abs);
  /* on ajoute autant de lignes que necessaire */
  while(line >= Nb_line)
    realloc_tab_mark_dart_priv(self);
  
  /* On positionne a 1 le bit qu'il faut dans l'octet */
  Element_tab[line][column] |= give_put_mask[b<0][mask];
}
void tab_mark_dart_unmark_dart(tab_mark_dart self,int b)
{
  int line,column,mask;
  dart b_abs;

  assert(self);
  
  b_abs = ABS(b);
  
  line   = NUM_LINE(b_abs);
  column = NUM_COLUMN(b_abs);
  mask   = NUM_MASK(b_abs);
  /* on ajoute autant de lignes que necessaire */
  while(line >= Nb_line)
    realloc_tab_mark_dart_priv(self);

  /* On met a zero le bit qu'il faut dans l'octet */
  Element_tab[line][column] &= (~give_put_mask[b<0][mask]);
}      
/*---------------------------------------------------------------------------*/
static four_mark* alloc_line_priv()
{
  four_mark* line;

  if(line_collector == NULL_GARBAGE)
    {
      line = (four_mark *) calloc(ALLOC_LINE,sizeof(four_mark));
      assert(line);
      return line;
    }
  if(garbage_is_empty(line_collector))
    {
      line = (four_mark *) calloc(ALLOC_LINE,sizeof(four_mark));
      assert(line);
      return line;
    }
  
  line =(four_mark*) garbage_give_object(line_collector);
  return line;
}
/*---------------------------------------------------------------------------*/
static tab_mark_dart alloc_tab_mark_dart_priv()
{
  tab_mark_dart self;
  
  if(tab_mark_dart_collector == NULL_GARBAGE)
    {
      self        = (tab_mark_dart) malloc(sizeof(struct tab_mark_dart));
      Element_tab = (four_mark **)malloc(ALLOC_NB_LINE*sizeof(four_mark*));
      Nb_of_pointer  = ALLOC_NB_LINE;
      
      return self;
    }
  if(garbage_is_empty(tab_mark_dart_collector))
    {
      self = (tab_mark_dart)malloc(sizeof(struct tab_mark_dart));
      Element_tab = (four_mark **)malloc(ALLOC_NB_LINE*sizeof(four_mark*));
      Nb_of_pointer  = ALLOC_NB_LINE;

      return self;
    }
  self = garbage_give_object(tab_mark_dart_collector);
  return self;
}
/*---------------------------------------------------------------------------*/
static void dealloc_line_priv(four_mark * line)
{
  four_mark *tmp;
  int i;

  if(line_collector == NULL_GARBAGE)
    {
      free(line);
      return;
    }
  tmp = line;
  for(i=0;i<ALLOC_LINE;i++)
    {
      *tmp = NULL_FOUR_MARK;
      tmp++;
    }
  garbage_insert_object(line_collector,line);
}
static void dealloc_tab_mark_dart_priv(tab_mark_dart self)
{
  if(tab_mark_dart_collector == NULL_GARBAGE)
    {
      free(self);
      return;
    }
  garbage_insert_object(tab_mark_dart_collector,self);
}
static void realloc_tab_mark_dart_priv(tab_mark_dart self)
{
  assert(self);

  /* On regarde si on na pas depasse le nb max de LINE alloue */
  if(Nb_line == Nb_of_pointer)
    {
      Element_tab = (four_mark**)realloc(Element_tab,
					   (Nb_of_pointer+ALLOC_NB_LINE)*
					   sizeof(four_mark*));
      assert(Element_tab);
      Nb_of_pointer += ALLOC_NB_LINE;
    }
  Element_tab[Nb_line] = alloc_line_priv();
  Nb_line++;
}
static void free_tab_mark_dart_priv(void* obj)
{
  tab_mark_dart self = obj;
  
  free(Element_tab);
  free(self);
}
  


