/* inverse_brun
 *  Copyright (C) 2000 Luc Brun
 *  email: brun@leri.univ-reims.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include <stdio.h>
#include <math.h>
#include <image.h>
#include <chrono.h>

#include "vect.h"
#include "covariance.h"
#include "image_dist.h"
#include "vect_propre.h"

#define UNKNOWN -1

#define NB_COLOR 5

static char* color_list_name[]=
{
  "rgb",
  "i1i2i3",
  "luv",
  "yiq",
  "k1k2k3"
};
static int color_list_flag[]=
{
  RGB,
  I1I2I3,
  LUV,
  YIQ,
  K1K2K3
};

#define NB_QUALITY 3

#define NORMAL   0
#define IMPROVED 1
#define BEST     2

static char* quality_name[]=
{
  "normal",
  "improved",
  "best"
};
static int quality_flag[]=
{
  NORMAL,
  IMPROVED,
  BEST
};

void usage(char* nom)
{
  fprintf(stderr,"USAGE :\n\n %s f1.ppm  tab.ppm [-space ",nom);
  fprintf(stderr,"rgb,k1k2k3,yiq,Luv,i1i2i3] ");
  fprintf(stderr,"[-quality normal improved best]\n");
}
void uniformiser(char* chaine)
{
  int i=0;
  char c;
  do
    {
      chaine[i] = tolower(chaine[i]);
      i++;
    }
  while(chaine[i] != '\0');

}
int choisir_argument(char* nom,
		     char* argument,
		     char* liste_valeur[],
		     int   liste_flag[],
		     int nb_valeur
		     )
{
  int i;
  
  uniformiser(argument);

  for(i=0;i<nb_valeur;i++)
    {
      if(strcmp(argument,liste_valeur[i])==0)
	return liste_flag[i];
    }
  
  usage(nom);
  exit(1);
}
int recuperer_argument(char* commande,
		       char* nom_param,
		       char* liste_valeur[],
		       int   liste_flag[],
		       int nb_valeur,
		       char* param[],
		       int* nb_param)
{
  int argument_choisi;
  int i,j;


  for(i=1;i<*nb_param;i++)
    if(strcmp(param[i],nom_param)==0)
      break;

  if(i==*nb_param)
    return UNKNOWN;
	  
  argument_choisi = choisir_argument(commande,				     
				     param[i+1],
				     liste_valeur,
				     liste_flag,
				     nb_valeur);
  
  for(j=i;j<*nb_param-2;j+=2)
    {
      param[j] = param[j+2];
      param[j+1] = param[j+3];
    }
  
  *nb_param = *nb_param-2;
  
  return argument_choisi;
}    
void affiche_results(mat_passage P)
{
  int i,j;
  vecteur v_propre;
  int coef_norm;
  double somme;
  
  for(i=0;i<DIM;i++)
    {
      
      if(P[1][i] <0)
	coef_norm = -1.0;
      else
	coef_norm = 1.0;
/*
      somme = 0.0;      
      for(j=0;j<DIM;j++)
	somme+=ABS(P[j][i]);
*/
      for(j=0;j<DIM;j++)
	v_propre[j] = coef_norm*P[j][i];
      /*      fprintf(stderr,"vecteur propre no %d :",i);
	      affiche_vect(v_propre);*/
    }
}
int main(int argc,char* argv[])
{
  char *nom;
  image table,img = FAIRE_image();
  mat_covariance mat;
  vecteur lambda;
  mat_passage P;
  int espace_choisi,qualite_choisi;
  double min1,min2,max1,max2;
  char* nom_table,*commande;
  double *table_proj;
  vecteur v1,v2;
  int max,i,nb_sources;
  image_dist img_dist;
  image img_res,img_dist_out;
  long temps_param,temps_calcul,temps_invers;

  ALLOC_list_gen();
  ALLOC_dart_list();
  ALLOC_tab_mark_dart();

  commande = argv[0];    
  espace_choisi = recuperer_argument(commande,
				     "-space",
				     color_list_name,
				     color_list_flag,
				     NB_COLOR,
				     argv,
				     &argc
				     );
  if(espace_choisi == UNKNOWN)
    espace_choisi = RGB;

  qualite_choisi = recuperer_argument(commande,
				      "-quality",
				      quality_name,
				      quality_flag,
				      NB_QUALITY,
				      argv,
				      &argc
				      );
  if(qualite_choisi == UNKNOWN)
    qualite_choisi = NORMAL;
    
  if( (argc != 3) )
    {
      usage(commande);
      exit(1);
    }
  nom = argv[1];
  nom_table = argv[2];
    
  if(image_charger(img,nom) == -1)
    {
      DEFAIRE_image(img);
      fprintf(stderr,"%s : Cannot access to %s\n",commande,nom);
      exit(2);
    }

  table = FAIRE_image();
  if(image_charger(table,nom_table) == -1)
    {
      DEFAIRE_image(table);
      fprintf(stderr,"%s : Cannot access to %s\n",commande,nom_table);
      exit(2);
    }
  nb_sources = image_give_largeur(table);
    
				/* Calcul de la matrice de covariance */
  temps_param = chrono(calcul_mat_cov(img,
				      mat,
				      espace_choisi));
				/* Calcul de la matrice de passage */
  temps_param += chrono(vecteur_valeur_propres(mat,lambda,P));



  /*    fprintf(stderr,"Matrice de covariance\n");
    affiche_mat(mat);*/
    
  /* fprintf(stderr,"lambda\n");
   affiche_vect(lambda);*/
    
    /*    fprintf(stderr,"Matrice de Passage\n");
	  affiche_results(P);			*/
			/* Extraction des deux premiers vecteurs  propres*/
  temps_param += chrono(
			deux_premiers_vecteurs_propres(P,v1,v2)
			);
  /* Calcul des projection minimums et maximum */
  temps_param += chrono( compute_max_min_proj(img,
					      v1,v2,
					      &min1,&max1,
					      &min2,&max2)
			 );
			  
  /*    max = sqrt(3.0)*256.0;
    min1 = -max;
    min2 = -max;
    max1 =  max;
    max2 =  max;*/
  /*
  fprintf(stderr,"min1 = %lf, max1 = %lf,max1-min1 =%lf\n",
	  min1,max1,max1-min1);
  fprintf(stderr,"min2 = %lf, max2 = %lf,max2-min2 = %lf\n",
	  min2,max2,max2-min2);*/
  /* Projection des couleurs representatives  dans le plan principal*/
  temps_param += chrono(
			table_proj = projection_table(table,v1,v2,min1,min2)
			);
    
  /*    for(i=0;i<2*nb_sources;i+=2)
      fprintf(stderr,"vect%d = (%lf,%lf)\n",i/2,table_proj[i],table_proj[i+1]);
  */  
  img_dist = FAIRE_image_dist();
    
  fprintf(stderr,"Temps de calcul des parametres :%f\n",
	  ((float)temps_param)/CLK_TCK);

  switch(qualite_choisi)
    {
    case NORMAL:

      temps_calcul = chrono(image_dist_initialize(img_dist,
						  max1-min1+1,
						  max2-min2+1,
						  table_proj,
						  nb_sources,
						  faux));
      temps_invers = chrono(img_res = inverse_colormap(img,
						       img_dist,
						       table,
						       v1,
						       min1,
						       v2,
						       min2
						       )
			    );
      break;
    case IMPROVED:
      temps_calcul = chrono(image_dist_initialize(img_dist,
						  max1-min1+1,
						  max2-min2+1,
						  table_proj,
						  nb_sources,
						  vrai)
			    );
      temps_invers = chrono(
			    img_res = slightly_improved_inverse_colormap(img,
									 img_dist,
									 table,
									 v1,
									 min1,
									 v2,
									 min2
									 )
			    );
      break;
    case BEST:
      temps_calcul = chrono(image_dist_initialize(img_dist,
						  max1-min1+1,
						  max2-min2+1,
						  table_proj,
						  nb_sources,
						  vrai)
			    );
      fprintf(stderr," valeur voisinage %f\n",
	      image_dist_mean_cardinal_of_delaunay_graph(img_dist));
      temps_invers = chrono(img_res = improved_inverse_colormap(img,
								img_dist,
								table,
								v1,
								min1,
								v2,
								min2
								)
			    );
      break;
    }
  fprintf(stderr,"Temps calcul img dist  :%f\n",
	  ((float)temps_calcul)/CLK_TCK);
  fprintf(stderr,"Temps d'inversion colormap  :%f\n",
	  ((float)temps_invers)/CLK_TCK);
  fprintf(stderr,"Temps total :%f\n",
	  ((float)(temps_param+temps_calcul+temps_invers))/CLK_TCK);
  image_to_stream(img_res,stdout);

  img_dist_out = image_dist_give_image(img_dist);
  image_sauvegarder(img_dist_out,"toto_dist.pgm");
  DEFAIRE_image(img_dist_out); 
    
  DEFAIRE_image_dist(img_dist);
  DEFAIRE_image(img_res);
}    
