/* inverse_brun
 *  Copyright (C) 2000 Luc Brun
 *  email: brun@leri.univ-reims.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include <stdio.h>
#include <math.h>
#include "covariance.h"
#include "macros.h"
#include "conversion.h"
#include "types.h"
#include "vect.h"
#include <image.h>

void affiche_vect(double*V)
{
    int i;
    
    printf("(");
    for(i=0;i<DIM-1;i++)
	    printf("%.2f, ",V[i]);
    printf("%.2f)\n",V[DIM-1]);
}

void affiche_mat(matrice M)
{
    int i ,j;
    
    for(i=0;i<DIM;i++)
	    {
		for(j=0;j<DIM;j++)
			printf("%f ",M[i][j]);
		printf("\n");
	    }
}
void entrer_mat(matrice M)
{
    int i;

    for(i=0;i<DIM;i++)
	    {
		LIRE_LIGNE(M,i);
	    }
}
double scalaire_vect(vecteur v1,vecteur v2)
{
  int i;
  double res;
  
  res = 0.0;
  for(i=0;i<DIM;i++)
    res += v1[i]*v2[i];
  
  return(res);
}
void vecteur_valeur_propres(mat_covariance a,vecteur val_propre,mat_passage P)
{

  vecteur x,y;
  double norme_x,norme_ymx,e,lambda;
  int i,j,k;
  int nb_vect_propre;

/* Ds les cas 2 et 3 des optimisation sont effectuees a la fin */
  if( (DIM ==2)||(DIM == 3) )
    nb_vect_propre = DIM-1;
  else
    nb_vect_propre = DIM;
  
  for(i=0;i<nb_vect_propre;i++)
    {
      for(j=0;j<DIM;j++)
	x[j] = j;
      do
	{
	  /*  Calcul de norme de x */
	  norme_x = scalaire_vect(x,x);
	  norme_x = sqrt(norme_x);
	  
	  /* calcul de y */
	  for(j=0;j<DIM;j++)
	    {
	      y[j]=0.0;
	      for(k=0;k<DIM;k++)
		y[j] += a[j][k]*x[k];
	      y[j] = y[j]/norme_x;
	    }
	  /* calcul de convergence */
	  norme_ymx =0.0;
	  for(j=0;j<DIM;j++)
	    norme_ymx +=(y[j]-x[j])*(y[j]-x[j]);
	  norme_ymx = sqrt(norme_ymx);
	  e = norme_ymx/norme_x;
	  
	  /* x = y*/
	  for(j=0;j<DIM;j++)
	    x[j] = y[j];			    	  
	}
      while(e > 0.0001);
      
      k=0;
      while((x[k] ==0)&&(k <DIM)) k++;

      norme_x = scalaire_vect(x,x);
      norme_x = sqrt(norme_x);

      for(j=0;j<DIM;j++)
	x[j] = x[j]/norme_x;
      
      if(k==DIM)
	{
	  fprintf(stderr,"PB vecteur propre nul\n");
	  exit(0);
	}
      else
	{
	  /* On normalise x */
	  
	  y[k]=0.0;
	  for(j=0;j<DIM;j++)				   
	    y[k] += a[k][j]*x[j];
	  
	  lambda = y[k]/x[k];
	}


      for(j=0;j<DIM;j++)
	for(k=0;k<DIM;k++)
	  a[j][k] -= lambda*x[k]*x[j];
      
      /* On stocke les resultats */
      for(j=0;j<DIM;j++)
	P[j][i] = x[j];
      
      val_propre[i] = lambda; 
		
    } /* Fin boucle principale */


  if( (DIM==2)||(DIM == 3) )
    {
      /* Calcul du dernier vecteur propre, val propre */
      if( DIM == 3 )
	{
	  /* La on utilise le fait que la matrice */
	  /* est 3*3 pour en deduire le dernier vecteur propre*/
	  /* Par un produit vectoriel */
	  for(i=0;i<DIM;i++)
	    {
	      j= (i+1)%DIM;
	      k= (i+2)%DIM;
	      P[i][2] = P[j][0]*P[k][1] - P[k][0]*P[j][1];
	    }
	}
      else
	{
	  if(DIM ==2)
	    {
	      P[0][1] = P[1][0];P[1][1] = -P[0][0];
	    }
	}
      norme_x = 0;
      for(j=0;j<DIM;j++)
	norme_x += P[j][DIM-1]*P[j][DIM-1];
      
      for(j=0;j<DIM;j++)
	P[j][DIM-1] = P[j][DIM-1]/norme_x;
      
      k=0;
      while((P[k][DIM-1] ==0)&&(k <DIM)) k++;
      
      if(k==DIM)
	{
	  fprintf(stderr,"PB vecteur propre nul\n");
	  exit(0);
	}
      else
	{
	  lambda = 0.0;
	  for(j=0;j<DIM;j++)
	    lambda += a[k][j]*P[j][DIM-1];
	  val_propre[DIM-1] = lambda/P[k][DIM-1];
	}
    }
}




void calcul_mat_cov(image img,mat_covariance M,int espace_choisi)
{
  int i,j,k,l;
  int R,G,B;
  int* tab;
  vecteur C;
  PFconversion *convert = changer_espace(espace_choisi); 
  double mom0=0.0;
  vecteur mom1={0.0,0.0,0.0};
  mat_covariance mom2 =
    {
      {0.0,0.0,0.0},
      {0.0,0.0,0.0},
      {0.0,0.0,0.0}
    };


  image_debut(img);
  do
    {
      tab = image_lire_pixel(img);
      
      R = tab[0];
      G = tab[1];
      B = tab[2];
      
      for(k=0;k<DIM;k++)
	C[k] = convert[k](R,G,B);
      
      mom0++;
      
      for(k=0;k<DIM;k++)
	mom1[k] += C[k];
      
      for(k=0;k<DIM;k++)
	for(l=0;l<DIM;l++)
	  mom2[k][l] += C[k]*C[l];
    }
  while(image_pixel_suivant(img));


  for(i=0;i<DIM;i++)
    for(j=0;j<=i;j++)
      M[i][j] =(mom2[i][j]-mom1[i]*(mom1[j]/mom0) )/mom0;
  for(i=0;i<DIM;i++)
    for(j=i+1;j<DIM;j++)
      M[i][j] = M[j][i];
}







