/* inverse_brun
 *  Copyright (C) 2000 Luc Brun
 *  email: brun@leri.univ-reims.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include <stdio.h>
#include <values.h>

#include <image.h>
#include <point.h>
#include <dart_list.h>

#include "vect.h"
#include "conversion.h"
#include "types.h"
#include "covariance.h"
#include "macros.h"
#include "image_dist.h"


#define MIN(a,b) ((a)<(b)?(a):(b))
#define MAX(a,b) ((a)>(b)?(a):(b))

#define MAX_TAILLE 512
typedef int colormap[MAX_TAILLE][3];

static int  nearest_color_priv(image_dist,
			       image ,
			       double,
			       double,
			       vecteur);

static void create_colormap_priv(image,
				 colormap,
				 int*);

void deux_premiers_vecteurs_propres(mat_passage P,
				    vecteur v1,
				    vecteur v2)
{

  int i;

  for(i=0;i<DIM;i++)
    v1[i] = P[i][0];
  
  for(i=0;i<DIM;i++)
    v2[i] = P[i][1];
}  
void compute_max_min_proj(image img,
			  vecteur v1,vecteur v2,
			  double *min_proj1,double* max_proj1,
			  double* min_proj2,double* max_proj2)
{
  int i,j;
  vecteur v;
  int* pixel;
  double min1,min2,max1,max2;
  double proj1,proj2;

  min1 = DBL_MAX;
  min2 = DBL_MAX;
  max1 = DBL_MIN;
  max2 = DBL_MIN;
  
  image_debut(img);
  do
    {
      pixel = image_lire_pixel(img);
      for(i=0;i<DIM;i++)
	v[i] = pixel[i];

      proj1 = scalaire_vect(v,v1);
      proj2 = scalaire_vect(v,v2);
      min1 = MIN(min1,proj1);
      min2 = MIN(min2,proj2);
      max1 = MAX(max1,proj1);
      max2 = MAX(max1,proj2);
    }
  while(image_pixel_suivant(img));

  *min_proj1 = min1;
  *min_proj2 = min2;
  *max_proj1 = max1;
  *max_proj2 = max2;
}
double *projection_table(image table,
			 vecteur v1,
			 vecteur v2,
			 double min1,
			 double min2)
{
  int hauteur = image_give_largeur(table);
  int *pixel;
  int i,j;
  vecteur v;
  double *proj = (double *)malloc(2*hauteur*sizeof(double));
  
  i=0;

  image_debut(table);
  do
    {
      pixel = image_lire_pixel(table);
      v[0] = pixel[0];
      v[1] = pixel[1];
      v[2] = pixel[2];

      proj[i] = scalaire_vect(v,v1)-min1;
      i++;
      proj[i] = scalaire_vect(v,v2)-min2;
      i++;
    }
  while(image_pixel_suivant(table));
  
  return proj;
}

image inverse_colormap(image img,
		       image_dist img_dist,
		       image table,
		       vecteur v1,
		       double min1,
		       vecteur v2,
		       double min2
		       )
{
  image img_res;
  int* pixel;
  int i,cell_nb;
  vecteur v;
  int x_proj,y_proj;
  Point p;
  colormap map;
  int nb_src;

  img_res = FAIRE_image();
  create_colormap_priv(table,map,&nb_src);
  image_initialize(img_res,
		   image_give_dim(img),
		   image_give_largeur(img),
		   image_give_hauteur(img)
		   );
  image_debut(img);
  image_debut(img_res);
  do
    {
      pixel = image_lire_pixel(img);

      x_proj = (int)( pixel[0]*v1[0] + 
		      pixel[1]*v1[1]  +
		      pixel[2]*v1[2]
		      -min1);

      y_proj = (int)( pixel[0]*v2[0] + 
		      pixel[1]*v2[1]  +
		      pixel[2]*v2[2]
		      -min2);
      
      cell_nb = image_dist_give_cell_nb(img_dist,x_proj,y_proj);
      image_ecrire_pixel(img_res,map[cell_nb]);

      image_pixel_suivant(img_res);
    } 
  while(image_pixel_suivant(img));
  
  return img_res;
}

image improved_inverse_colormap(image img,
				image_dist img_dist,
				image table,
				vecteur v1,
				double min1,
				vecteur v2,
				double min2
				)
{ 
  image img_res;
  int* pixel;
  int i,cell_nb;
  vecteur v;
  int x_proj,y_proj;
  Point p;
  
  img_res = FAIRE_image();
  image_initialize(img_res,
		   image_give_dim(img),
		   image_give_largeur(img),
		   image_give_hauteur(img)
		   );
  image_debut(img);
  image_debut(img_res);
  do
    {
      pixel = image_lire_pixel(img);
      for(i=0;i<3;i++)
	v[i] = (double)pixel[i];
      
      x_proj = (int)(scalaire_vect(v,v1)-min1);
      y_proj = (int)(scalaire_vect(v,v2)-min2);
      cell_nb = nearest_color_priv(img_dist,table,x_proj,y_proj,v);
      COORDX(p) = cell_nb;
      COORDY(p) = 0;
      image_move_to(table,&p);
      pixel = image_lire_pixel(table);
	
      image_ecrire_pixel(img_res,pixel);
      image_pixel_suivant(img_res);
    } 
  while(image_pixel_suivant(img));
  
  return img_res;
}
image slightly_improved_inverse_colormap(image img,
					 image_dist img_dist,
					 image table,
					 vecteur v1,
					 double min1,
					 vecteur v2,
					 double min2
					 )
{ 
  image img_res;
  int* pixel;
  int i,cell_nb;
  vecteur v;
  int x_proj,y_proj;
  Point p;

  img_res = FAIRE_image();
  image_initialize(img_res,
		   image_give_dim(img),
		   image_give_largeur(img),
		   image_give_hauteur(img)
		   );
  image_debut(img);
  image_debut(img_res);
  do
    {
      pixel = image_lire_pixel(img);
      for(i=0;i<3;i++)
	v[i] = (double)pixel[i];
      
      x_proj = (int)(scalaire_vect(v,v1)-min1);
      y_proj = (int)(scalaire_vect(v,v2)-min2);
      
      if(image_dist_is_point_frontier(img_dist,x_proj,y_proj))
	cell_nb = nearest_color_priv(img_dist,table,x_proj,y_proj,v);
      else
	cell_nb  = image_dist_give_cell_nb(img_dist,x_proj,y_proj);
      
      COORDX(p) = cell_nb;
      COORDY(p) = 0;
      image_move_to(table,&p);
      pixel = image_lire_pixel(table);
      
      image_ecrire_pixel(img_res,pixel);
      image_pixel_suivant(img_res);
    } 
  while(image_pixel_suivant(img));
  
  return img_res;
}
  

static int  nearest_color_priv(image_dist img_dist,
			       image table,
			       double x_proj,
			       double y_proj,
			       vecteur pixel)
{
  int cell_nb,cell;
  Point p;
  dart_list adj_list;
  int *rep_color;
  int nearest_color;
  int dist,min_dist;
  
  cell_nb  = image_dist_give_cell_nb(img_dist,x_proj,y_proj);
  adj_list = image_dist_give_adjacent_cells(img_dist,cell_nb);
  COORDX(p) = cell_nb;
  COORDY(p) = 0;
  image_move_to(table,&p);
  rep_color = image_lire_pixel(table);
  min_dist  = ((rep_color[0]-pixel[0])*(rep_color[0]-pixel[0]) +
	       (rep_color[1]-pixel[1])*(rep_color[1]-pixel[1])  +
	       (rep_color[2]-pixel[2])*(rep_color[2]-pixel[2])
	       );
  nearest_color = cell_nb;
  dart_list_first(adj_list);
  do
    {
      cell = dart_list_give_current(adj_list);
      COORDX(p) = cell;
      image_move_to(table,&p);
      rep_color = image_lire_pixel(table);
      dist = ((rep_color[0]-pixel[0])*(rep_color[0]-pixel[0]) +
	      (rep_color[1]-pixel[1])*(rep_color[1]-pixel[1])  +
	      (rep_color[2]-pixel[2])*(rep_color[2]-pixel[2])
	      );
      if(dist < min_dist)
	{
	  nearest_color = cell;
	  min_dist = dist;
	}
    }
  while(dart_list_next(adj_list));

  return nearest_color;
}
static void create_colormap_priv(image table,
				 colormap map,
				 int* nb_entry)
{
  int nb_color=0;
  int *pixel;
  int i;
 
  image_debut(table);
  do
    {
      pixel = image_lire_pixel(table);
      for(i=0;i<3;i++)
	map[nb_color][i] = pixel[i];
      nb_color++;
    }
  while(image_pixel_suivant(table));
  
  *nb_entry = nb_color;
}
